# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

from rich.style import Style
from rich.text import Text, TextType

RE_CVE = r"CVE\-\d{4}\-\d+"


def linkify_cve(text_str: TextType) -> Text:
    """Apply a link to anything that looks like a CVE."""

    def make_link(cve: str) -> Style:
        return Style(link=f"https://nvd.nist.gov/vuln/detail/{cve}")

    text = Text.from_markup(text_str) if isinstance(text_str, str) else text_str
    text.highlight_regex(RE_CVE, style=make_link)
    return text



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Retrieval access and of NVD entries using NVD Automatic CVE Retrieval

Parameter values and more information: https://nvd.nist.gov/developers/products
"""

import asyncio
import json
import math
import time
from datetime import datetime, timedelta, timezone
from logging import Logger
from typing import Dict, List, Union

import aiohttp
from rich.progress import Progress, track

from cve_ape.async_utils import RateLimiter
from cve_ape.error_handler import ErrorMode, NVDKeyError, NVDServiceError
from cve_ape.log import LOGGER

FEED = "https://services.nvd.nist.gov/rest/json/cves/1.0"
NVD_CVE_STATUS = "https://nvd.nist.gov/rest/public/dashboard/statistics"

PAGESIZE = 2000
MAX_FAIL = 5
# Interval in seconds between successive requests
INTERVAL_PERIOD = 3


class NVD_API:
    def __init__(
        self,
        logger: Logger = LOGGER.getChild("NVD_API"),
        feed=FEED,
        session=None,
        page_size: int = PAGESIZE,
        max_fail: int = MAX_FAIL,
        interval: int = INTERVAL_PERIOD,
        error_mode: ErrorMode = ErrorMode.TruncTrace,
        incremental_update=False,
        api_key: str = "",
    ):
        self.logger = logger or LOGGER.getChild(self.__class__.__name__)
        self.feed = feed
        self.session = session
        self.params: Dict = dict()
        self.page_size = page_size
        self.max_fail = max_fail
        self.interval = interval
        self.error_mode = error_mode
        self.incremental_update = incremental_update
        self.total_results = -1
        self.failed_count = 0
        self.all_cve_entries: List = []
        if api_key:
            self.params["apiKey"] = api_key

    @staticmethod
    def convert_date_to_nvd_date(date: datetime) -> str:
        """Returns a datetime string of NVD recognized date format"""
        utc_date = date.astimezone(timezone.utc).strftime("%Y-%m-%dT%H:%M:%S:%f")[:-3]
        return f"{utc_date} UTC-00:00"

    @staticmethod
    async def nvd_count_metadata(session):
        """Returns CVE Status count from NVD"""
        cve_count = {
            "Total": 0,
            "Rejected": 0,
            "Received": 0,
        }
        async with await session.get(
            NVD_CVE_STATUS,
            params={"reporttype": "countsbystatus"},
            raise_for_status=True,
        ) as response:
            data = await response.json()
            for key in data["vulnsByStatusCounts"]:
                cve_count[key["name"]] = int(key["count"])
        return cve_count

    async def get_nvd_params(
        self,
        time_of_last_update: Union[datetime, None] = None,
    ):
        """
        Initialize NVD request parameters
        """
        self.params["startIndex"] = 0
        self.params["resultsPerPage"] = self.page_size

        if not self.session:
            connector = aiohttp.TCPConnector(limit_per_host=19)
            self.session = RateLimiter(
                aiohttp.ClientSession(connector=connector, trust_env=True)
            )

        self.logger.debug("Fetching metadata from NVD...")
        cve_count = await self.nvd_count_metadata(self.session)

        if "apiKey" in self.params:
            await self.validate_nvd_api()

        if time_of_last_update:
            # Fetch all the updated CVE entries from the modified date. Subtracting 2-minute offset for updating cve entries
            self.params["modStartDate"] = self.convert_date_to_nvd_date(
                time_of_last_update - timedelta(minutes=2)
            )
            self.params["modEndDate"] = self.convert_date_to_nvd_date(datetime.now())
            self.logger.info(
                f'Fetching updated CVE entries after {self.params["modStartDate"]}'
            )
            self.params["includeMatchStringChange"] = json.dumps(True)
            # Check modified strings inside CVEs as well
            with Progress() as progress:
                task = progress.add_task(
                    "Fetching incremental metadata from NVD...", total=1, start=False
                )
            while await self.load_nvd_request(start_index=0):
                progress.update(task)
            progress.update(task, advance=1)

        else:
            self.total_results = cve_count["Total"] - cve_count["Rejected"]
        self.logger.info(f"Adding {self.total_results} CVE entries")

    async def validate_nvd_api(self):
        """
        Validate NVD API
        """
        param_dict = self.params.copy()
        param_dict["startIndex"] = 0
        param_dict["resultsPerPage"] = 1
        try:
            self.logger.debug("Validating NVD API...")
            async with await self.session.get(
                self.feed, params=param_dict, raise_for_status=True
            ) as response:
                data = await response.json()
                if data.get("error", False):
                    self.logger.error(f"NVD API error: {data['error']}")
                    raise NVDKeyError(self.params["apiKey"])
        except NVDKeyError:
            # If the API key provided is invalid, delete from params
            # list and try the request again.
            self.logger.error("unset api key, retrying")
            del self.params["apiKey"]

    async def load_nvd_request(self, start_index):
        """Get single NVD request and update year_wise_data list which contains list of all CVEs"""

        param_dict = self.params.copy()
        param_dict["startIndex"] = start_index

        fetched_data = None
        while fetched_data is None:
            try:
                async with await self.session.get(
                    self.feed,
                    params=param_dict,
                    raise_for_status=True,
                ) as response:
                    if response.status == 200:
                        fetched_data = await response.json()

                        if start_index == 0:
                            # Update total results in case there is discrepancy between NVD dashboard and API
                            self.total_results = fetched_data["totalResults"]
                        self.all_cve_entries.extend(fetched_data["result"]["CVE_Items"])

                    elif response.status == 503:
                        raise NVDServiceError(self.params["modStartDate"])
                    else:
                        self.failed_count += 1
                        if self.failed_count == self.max_fail:
                            self.failed_count = 0
                            self.logger.info(
                                f"Pausing requests for {self.interval} seconds"
                            )
                            time.sleep(self.interval)
                        else:
                            time.sleep(1)

            except Exception as error:
                self.logger.debug(f"Failed to connect to NVD {error}")
                self.logger.debug(f"Pausing requests for {self.interval} seconds")
                self.failed_count += 1
                time.sleep(self.interval)

    async def get(self):
        """Calls load_nvd_request() multiple times to fetch all NVD feeds"""
        start_index = 1 if self.incremental_update else 0
        nvd_requests = [
            self.load_nvd_request(index * self.page_size)
            for index in range(
                start_index, 1 + int(math.ceil(self.total_results / self.page_size))
            )
        ]
        total_tasks = len(nvd_requests)
        # error_mode.value will only be greater than 1 if quiet mode.
        if self.error_mode.value > 1:
            iter_tasks = track(
                asyncio.as_completed(nvd_requests),
                description="Downloading Feeds from NVD...",
                total=total_tasks,
            )
        else:
            iter_tasks = asyncio.as_completed(nvd_requests)

        for task in iter_tasks:
            await task



import csv
import re
from collections import defaultdict
from logging import Logger
from os.path import dirname, getsize, isfile, join
from csv import DictReader, reader, unix_dialect

from cve_ape.cvedb import CVEDB
from cve_ape.error_handler import (
    EmptyTxtError,
    ErrorHandler,
    ErrorMode,
)
from cve_ape.log import LOGGER
from cve_ape.util import ProductInfo, Remarks

ROOT_PATH = join(dirname(__file__), "..")


class PackageListParser:
    def __init__(
        self, input_file: str, logger: Logger = None, csv_format:str = None, csv_delimiter:str = ";", error_mode=ErrorMode.TruncTrace
    ) -> None:
        self.input_file = input_file
        self.logger = logger or LOGGER.getChild(self.__class__.__name__)
        self.error_mode = error_mode
        self.csv_format = csv_format
        self.csv_delimiter = csv_delimiter
        self.parsed_data_without_vendor = defaultdict(dict)
        self.parsed_data_with_vendor = defaultdict(dict)
        self.package_names_known = []
        self.package_names_unknown = []

    def process_file(self, input_file):
        packages = []
        with open(input_file) as req:
            lines = req.readlines()
        parser_re = re.compile(r"^(.*?)\-(?=(?:[0-9]+\.){0,}[0-9]+(?:-[a-z]+)?)(.*?)$")
        for line in lines:
            parse = parser_re.search(re.split("\n", line)[0])
            if parse is not None:
                packages.append({"name": parse[1], "version": parse[2]})
            else:
                self.logger.warning(f"Did not parsed: {line}")
        return packages

    def process_file_csv(self, input_file, fields):
        packages = []
        dialect = unix_dialect
        dialect.delimiter = self.csv_delimiter
        with open(input_file) as read_obj:
            if fields[0].isnumeric() and fields[1].isnumeric():
                rd = reader(read_obj, dialect=dialect)
                list_out = list(rd)
                for line in list_out:
                    if len(line) >= 2:
                        packages.append({
                            "name": line[int(fields[0])],
                            "version": line[int(fields[1])]
                            })
            else:
                rd = DictReader(read_obj, dialect=dialect)
                list_out = list(rd)
                for line in list_out:
                    keys = line.keys()
                    if fields[0] in keys and fields[1] in keys:
                        packages.append({"name": line[fields[0]], "version": line[fields[1]]})
        return packages

    def parse_list(self):
        input_file = self.input_file
        file_type = self.check_file()
        if file_type == 'csv':
            if self.csv_format is not None:
                fields = self.csv_format.split(",")
                if len(fields) != 2:
                    self.logger.warning("CSV fields were not recognized. This is what I get: {}. "
                                      "Trying to use a default settings.".format(fields))
                    packages = self.process_file_csv(input_file, ("Name", "Version"))
                else:
                    packages = self.process_file_csv(input_file, tuple(fields))
            else:
                packages = self.process_file_csv(input_file, ("Name", "Version"))
        else:
            if self.csv_format is not None:
                self.logger.warning("CSV file is not recognized. Trying to process as a txt file.")
            packages = self.process_file(input_file)
        cve_db = CVEDB()
        self.package_names_known, self.package_names_unknown = cve_db.get_packages_from_pkglist(
            packages
        )
        return self.package_names_known

    def parse_data(self):
        for row in self.package_names_known:
            product_info = ProductInfo(
                row["vendor"], row["name"].lower(), row["version"]
            )
            self.parsed_data_with_vendor[product_info][
                row.get("cve_number", "").strip() or "default"
            ] = {
                "remarks": Remarks.NewFound,
                "comments": row.get("comments", "").strip(),
                "severity": row.get("severity", "").strip(),
            }
            self.parsed_data_with_vendor[product_info]["paths"] = {""}

    def check_file(self):
        input_file = self.input_file
        error_mode = self.error_mode

        if not isfile(input_file):
            with ErrorHandler(mode=error_mode):
                raise FileNotFoundError(input_file)

        if getsize(input_file) == 0:
            with ErrorHandler(mode=error_mode):
                raise EmptyTxtError(input_file)

        with open(input_file) as f:
            first_line = f.readline()
            if re.match(r'^[^\;\,]+$', first_line):
                return 'txt'
            else:
                return 'csv'



import argparse
import logging
import os
import sys
import textwrap
import time
from collections import ChainMap

from cve_ape.cve_scanner import CVEScanner
from cve_ape.cvedb import CVEDB, OLD_CACHE_DIR, DISK_LOCATION_DEFAULT
from cve_ape.error_handler import (
    CVEDataMissing,
    EmptyCache,
    ErrorHandler,
    ErrorMode,
    InsufficientArgs,
    excepthook,
)
from cve_ape.log import LOGGER
from cve_ape.output_engine import OutputEngine
from cve_ape.package_list_parser import PackageListParser

if sys.version_info >= (3, 8):
    pass
else:
    pass

sys.excepthook = excepthook  # Always install excepthook for entrypoint module.


class StringToListAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        values = list(map(lambda val: val.strip(), values.split(",")))
        setattr(namespace, self.dest, values)


def main(argv=None):
    """Scan a binary file for certain open source libraries that may have CVEs"""
    argv = argv or sys.argv

    # Reset logger level to info
    LOGGER.setLevel(logging.INFO)

    parser = argparse.ArgumentParser(
        prog="cve-ape",
        description=textwrap.dedent(
            """
            """
        ),
        epilog=textwrap.fill(
            f''
        ),
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    nvd_database_group = parser.add_argument_group(
        "CVE Data Download", "Arguments related to NVD Database and Cache Configuration"
    )
    nvd_database_group.add_argument(
        "-n",
        "--nvd",
        action="store",
        choices=["api", "json"],
        help="choose method for getting CVE lists from NVD",
        default="api",
    )
    nvd_database_group.add_argument(
        "-u",
        "--update",
        action="store",
        choices=["now", "daily", "never", "latest"],
        help="update schedule for NVD database (default: daily)",
        default="daily",
    )
    nvd_database_group.add_argument(
        "--cache-dir",
        action="store",
        default=DISK_LOCATION_DEFAULT,
        help="specify the NVD database and cache location (default: ~/.cache/cve-ape)",
    )
    nvd_database_group.add_argument(
        "--nvd-api-key",
        action="store",
        default="",
        help="specify NVD API key (used to improve NVD rate limit)",
    )

    input_group = parser.add_argument_group("Input")
    input_group.add_argument(
        "-L", "--package-list", action="store", default="", help="provide a package list."
    )
    input_group.add_argument(
        "-C", "--csv-format", action="store", default="", help="comma separated headers if file is a CSV, "
                                                               "or comma separated column numbers started from 0 where the "
                                                               "first is a package and the second is a version."
    )

    input_group.add_argument(
        "-d", "--csv-delimiter", action="store", default=";", help="a delimiter used in the csv file. Default: \";\""
    )

    output_group = parser.add_argument_group("Output")
    output_group.add_argument(
        "-q",
        "--quiet",
        action="store_true",
        help="suppress output",
        default=False,
    )
    output_group.add_argument(
        "-l",
        "--log",
        help="log level (default: info)",
        dest="log_level",
        action="store",
        choices=["debug", "info", "warning", "error", "critical"],
        default="info",
    )
    output_group.add_argument(
        "-o",
        "--output-file",
        action="store",
        help="provide output filename (default: output to stdout)",
        default="",
    )

    output_group.add_argument(
        "-f",
        "--format",
        action="store",
        choices=["csv", "json", "console"],
        help="update output format (default: console)",
        default="console",
    )
    output_group.add_argument(
        "-c",
        "--cvss",
        action="store",
        help="minimum CVSS score (as integer in range 0 to 10) to report (default: 0)",
        default=0,
    )
    output_group.add_argument(
        "-S",
        "--severity",
        action="store",
        choices=["low", "medium", "high", "critical"],
        help="minimum CVE severity to report (default: low)",
        default="low",
    )
    parser.add_argument(
        "-e",
        "--exclude",
        action=StringToListAction,
        help="Line separated Exclude package list. It shell be the same format as an inclusion list."
             " e.g. if you use CSV to test, you should use CSV for exclusion too",
        default=[],
    )
    parser.add_argument(
        "--disable-version-check",
        action="store_true",
        help="skips checking for a new version",
        default=False,
    )
    parser.add_argument(
        "--offline",
        action="store_true",
        help="operate in offline mode",
        default=False,
    )

    with ErrorHandler(mode=ErrorMode.NoTrace):
        raw_args = parser.parse_args(argv[1:])
        args = {key: value for key, value in vars(raw_args).items() if value}
        defaults = {key: parser.get_default(key) for key in vars(raw_args)}

    configs = {}

    args = ChainMap(args, configs, defaults)

    # logging and error related settings
    if args["log_level"]:
        LOGGER.setLevel(args["log_level"].upper())

    if args["quiet"]:
        LOGGER.setLevel(logging.CRITICAL)

    if 0 < LOGGER.level <= 10:
        error_mode = ErrorMode.FullTrace
    elif LOGGER.level >= 50:
        error_mode = ErrorMode.NoTrace
    else:
        error_mode = ErrorMode.TruncTrace

    # once logging is set, we can output the version and NVD notice
    LOGGER.info(f"Package list CVE search tool v1.0")
    LOGGER.info(
        "This product uses the NVD API but is not endorsed or certified by the NVD."
    )

    # If NVD API key is not set, check for environment variable (e.g. GitHub Secrets)
    if not args["nvd_api_key"] and os.getenv("nvd_api_key"):
        args["nvd_api_key"] = os.getenv("nvd_api_key")

    # Also try the uppercase env variable, in case people prefer those
    if not args["nvd_api_key"] and os.getenv("NVD_API_KEY"):
        args["nvd_api_key"] = os.getenv("NVD_API_KEY")

    # If you're not using an NVD key, let you know how to get one
    if not args["nvd_api_key"] and not args["offline"]:
        LOGGER.info("Not using an NVD API key. Your access may be rate limited by NVD.")
        LOGGER.info(
            "Get an NVD API key here: https://nvd.nist.gov/developers/request-an-api-key"
        )

    # CSVScanner related settings
    score = 0
    if args["severity"]:
        # Set minimum CVSS score based on severity
        cvss_score = {"low": 0, "medium": 4, "high": 7, "critical": 9}
        score = cvss_score[args["severity"]]
    if int(args["cvss"]) > 0:
        score = int(args["cvss"])

    # Offline processing
    if args["offline"]:
        # Override version check and database update arguments
        version_check = True
        db_update = "never"
    else:
        version_check = args["disable_version_check"]
        db_update = args["update"]

    output_format = args["format"]

    # Database update related settings
    # Connect to the database
    cvedb_orig = CVEDB(
        cachedir=os.path.expanduser(args["cache_dir"]),
        version_check=not version_check,
        error_mode=error_mode,
        nvd_type=args["nvd"],
        incremental_update=True if db_update == "latest" and args["nvd"] else False,
        nvd_api_key=args["nvd_api_key"],
    )

    # if OLD_CACHE_DIR (from cvedb.py) exists, print warning
    if os.path.exists(OLD_CACHE_DIR):
        LOGGER.warning(
            f"Obsolete cache dir {OLD_CACHE_DIR} is no longer needed and can be removed."
        )

    # Check database exists if operating in offline mode.
    if args["offline"] and not cvedb_orig.check_db_exists():
        LOGGER.critical("Database does not exist. Make sure, that you've run the tool at least on online.")
        return -1

    # Clear data if -u now is set
    if db_update == "now":
        cvedb_orig.clear_cached_data()

    if db_update == "latest":
        cvedb_orig.refresh_cache_and_update_db()

    # update db if needed
    if db_update != "never":
        cvedb_orig.get_cvelist_if_stale()
    else:
        if args["nvd"] == "json":
            LOGGER.warning("Not verifying CVE DB cache")
            cvedb_orig.get_db_update_date()
            if not cvedb_orig.nvd_years():
                with ErrorHandler(mode=error_mode, logger=LOGGER):
                    raise EmptyCache(cvedb_orig.cachedir)

    # CVE Database validation
    if not cvedb_orig.check_cve_entries():
        with ErrorHandler(mode=error_mode, logger=LOGGER):
            raise CVEDataMissing("No data in CVE Database")

    # Report time of last database update
    db_date = time.strftime(
        "%d %B %Y at %H:%M:%S", time.localtime(cvedb_orig.get_db_update_date())
    )
    LOGGER.info(f"CVE database last updated on {db_date}")

    cvedb_orig.remove_cache_backup()

    # Input validation
    if (
        not args["package_list"]
    ):
        parser.print_usage()
        with ErrorHandler(logger=LOGGER, mode=ErrorMode.NoTrace):
            raise InsufficientArgs(
                "A package list file required"
            )

    with CVEScanner(score=score) as cve_scanner:
        # Package List parsing
        if args["package_list"]:
            if args["csv_format"]:
                package_list = PackageListParser(
                    args["package_list"], csv_format=args["csv_format"], csv_delimiter=args["csv_delimiter"], error_mode=error_mode
                )
            else:
                package_list = PackageListParser(
                    args["package_list"], error_mode=error_mode
                )
            parsed_data = package_list.parse_list()
            cves = cve_scanner.get_cves(parsed_data)

        LOGGER.info("")
        LOGGER.info("Overall CVE summary: ")

        if cves:
            LOGGER.info(f"Known CVEs: {len(cves.keys())}:")

            # Creates an Object for OutputEngine
            output = OutputEngine(
                all_cve_data=cves,
                filename=args["output_file"],
            )

            if not args["quiet"]:
                output.output_file(output_format)

        # If no cves found, then the program exits cleanly (0 exit)
        if cve_scanner.products_with_cve == 0:
            return 0

        # if some cves are found, return with exit code 1
        # Previously this returned a number of CVEs found, but that can
        # exceed expected return value range.
        if cve_scanner.products_with_cve > 0:
            return 1

        # If somehow we got negative numbers of cves something has gone
        # horribly wrong.  Since return code 2 is used by argparse, use 3
        return 3


if __name__ == "__main__":
    if os.getenv("NO_EXIT_CVE_NUM"):
        main()
    else:
        sys.exit(main())



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

# This file also includes the RateLimiter function with the following license:
#
#    Copyright 2018 Quentin Pradet
#
#    Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

#    The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

#    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


""" Utility classes for the CVE Binary Tool """

import asyncio
import glob
import gzip
import itertools
import os
import shutil
import subprocess
import sys
import tempfile
import time
from functools import partial, wraps

from cve_ape.util import inpath


def async_wrap(func):
    @wraps(func)
    async def run(*args, loop=None, executor=None, **kwargs):
        if loop is None:
            loop = asyncio.get_event_loop()
        pfunc = partial(func, *args, **kwargs)
        return await loop.run_in_executor(executor, pfunc)

    return run


def get_event_loop():
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    if sys.platform.startswith("win"):
        if isinstance(loop, asyncio.SelectorEventLoop):
            loop = asyncio.ProactorEventLoop()
            asyncio.set_event_loop(loop)
    return loop


def run_coroutine(coro):
    loop = get_event_loop()
    aws = asyncio.ensure_future(coro, loop=loop)
    result = loop.run_until_complete(aws)
    return result


async def aio_run_command(args, process_can_fail=True):
    process = await asyncio.create_subprocess_exec(
        *args,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    stdout, stderr = await process.communicate()
    if process.returncode != 0 and not process_can_fail:
        raise subprocess.CalledProcessError(
            args, process.returncode, output=stdout, stderr=stderr
        )
    return stdout, stderr, process.returncode  # binary encoded


class ChangeDirContext:
    def __init__(self, destination_dir):
        self.current_dir = os.getcwd()
        self.destination_dir = destination_dir

    async def __aenter__(self):
        os.chdir(self.destination_dir)

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self.current_dir)


class FileIO:
    _open = async_wrap(open)
    _name_idx = 0
    _mode_idx = 1
    _mode = "r"

    def __init__(self, *args, **kwargs):
        # Do some trick to get exact filename and mode regardless of args or kwargs
        flatargs = list(itertools.chain(args, kwargs.values()))
        if (
            self.__class__._name_idx is not None
            and len(flatargs) > self.__class__._name_idx
        ):
            self.name = flatargs[self.__class__._name_idx]
        if (
            self.__class__._mode_idx is not None
            and len(flatargs) > self.__class__._mode_idx
        ):
            self._mode = flatargs[self.__class__._mode_idx]
        self._args = args
        self._kwargs = kwargs
        self._file = None

    async def __call__(self):
        """Convenience method to allow call like following:
        f = await FileIO("some file path", "r")()
        Note: We can't make async __init__
        """
        return await self.open()

    async def open(self):
        file = await self.__class__._open(*self._args, **self._kwargs)
        self._file = file
        self._setup()
        return self

    def _setup(self):
        if not self._file:
            raise RuntimeError("Invalid Use: Call open() before calling _setup()")
        common_async_attrs = {
            "close",
            "flush",
            "isatty",
            "read",
            "readline",
            "readlines",
            "seek",
            "tell",
            "truncate",
            "write",
            "writelines",
        }
        common_sync_attrs = {
            "detach",
            "fileno",
            "readable",
            "writable",
            "seekable",
            "closed",
            "mode",
            "name",
        }
        strings_sync_attrs = {
            "buffer",
            "encoding",
            "errors",
            "line_buffering",
            "newlines",
        }
        bytes_async_attrs = {"readinto", "readinto1", "read1"}
        bytes_sync_attrs = {"raw"}
        if "b" in self._mode:
            async_attrs = common_async_attrs | bytes_async_attrs
            sync_attrs = common_sync_attrs | bytes_sync_attrs
        else:
            async_attrs = common_async_attrs
            sync_attrs = common_sync_attrs | strings_sync_attrs
        [
            setattr(self, attr, async_wrap(getattr(self._file, attr)))
            for attr in async_attrs
            if hasattr(self._file, attr)
        ]
        [
            setattr(self, attr, getattr(self._file, attr))
            for attr in sync_attrs
            if hasattr(self._file, attr)
        ]

    async def __aenter__(self):
        return await self.open()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self.close()

    async def __anext__(self):
        line = await self.readline()
        if line:
            return line
        else:
            raise StopAsyncIteration

    def __aiter__(self):
        return self


class TemporaryFile(FileIO):
    _open = async_wrap(tempfile.TemporaryFile)
    _name_idx = None
    _mode_idx = 0
    _mode = "w+b"

    def _setup(self):
        super()._setup()
        self.name = self._file.name


class NamedTemporaryFile(TemporaryFile):
    _open = async_wrap(tempfile.NamedTemporaryFile)


class SpooledTemporaryFile(TemporaryFile):
    _open = async_wrap(tempfile.SpooledTemporaryFile)
    _mode_idx = 1


class GzipFile(FileIO):
    _open = async_wrap(gzip.GzipFile)


class RateLimiter:
    """Rate limits an HTTP client that would make get() and post() calls.
    Calls are rate-limited by host.
    https://quentin.pradet.me/blog/how-do-you-rate-limit-calls-with-aiohttp.html
    This class is not thread-safe.

    Copyright 2018 Quentin Pradet
    See license at top of file.
    """

    RATE = 10
    MAX_TOKENS = 10

    def __init__(self, client):
        self.client = client
        self.tokens = self.MAX_TOKENS
        self.updated_at = time.monotonic()

    async def get(self, *args, **kwargs):
        await self.wait_for_token()
        return self.client.get(*args, **kwargs)

    async def wait_for_token(self):
        while self.tokens < 1:
            self.add_new_tokens()
            await asyncio.sleep(0.1)
        self.tokens -= 1

    def add_new_tokens(self):
        now = time.monotonic()
        time_since_update = now - self.updated_at
        new_tokens = time_since_update * self.RATE
        if self.tokens + new_tokens >= 1:
            self.tokens = min(self.tokens + new_tokens, self.MAX_TOKENS)
            self.updated_at = now

    async def close(self):
        await self.client.close()


aio_rmdir = async_wrap(shutil.rmtree)
aio_rmfile = async_wrap(os.remove)
aio_unpack_archive = async_wrap(shutil.unpack_archive)
aio_glob = async_wrap(glob.glob)
aio_mkdtemp = async_wrap(tempfile.mkdtemp)
aio_makedirs = async_wrap(os.makedirs)
aio_inpath = async_wrap(inpath)



import textwrap
from collections import defaultdict
from datetime import datetime

from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel
from rich.table import Table
from rich.text import Text

from .util import get_cve_summary
from ..linkify import linkify_cve
from ..theme import cve_theme
from ..util import VersionInfo

SUMMARY_COLOR = {
    "CRITICAL": "red",
    "HIGH": "blue",
    "MEDIUM": "yellow",
    "LOW": "green",
}

SUMMARY_ORDER = {
    "CRITICAL": 1,
    "HIGH": 2,
    "MEDIUM": 3,
    "LOW": 4,
}


def format_version_range(version_info: VersionInfo) -> str:
    (start_including, start_excluding, end_including, end_excluding, is_conditional) = version_info
    if start_including and end_including:
        return f"[{start_including} - {end_including}]"
    if start_including and end_excluding:
        return f"[{start_including} - {end_excluding})"
    if start_excluding and end_including:
        return f"({start_excluding} - {end_including}]"
    if start_excluding and end_excluding:
        return f"({start_excluding} - {end_excluding})"
    if start_including:
        return f">= {start_including}"
    if start_excluding:
        return f"> {start_excluding}"
    if end_including:
        return f"<= {end_including}"
    if end_excluding:
        return f"< {end_excluding}"
    return "-"


def score_colors(input_score):
    if type(input_score) == int or type(input_score) == float:
        if 0 <= input_score < 3.9:
            return "white"
        elif 3.9 <= input_score < 5.9:
            return "yellow"
        elif 5.9 <= input_score < 7.9:
            return "red"
        elif input_score >= 7.9:
            return "black"
    else:
        try:
            return SUMMARY_COLOR[input_score]
        except:
            return "white"


def output_console(
        all_cve_data: {},
        console=Console(theme=cve_theme),
):
    console._width = 120
    now = datetime.now().strftime("%Y-%m-%d  %H:%M:%S")

    console.print(
        Markdown(
            textwrap.dedent(
                f"""
                Report Generated: {now}
                """
            )
        )
    )

    # Create table instance for CVE Summary
    table = Table()
    # Add Head Columns to the Table
    table.add_column("Severity")
    table.add_column("Count")
    summary = get_cve_summary(all_cve_data)

    for severity, count in summary.items():
        color = SUMMARY_COLOR[severity]
        cells = [
            Text.styled(severity, color),
            Text.styled(str(count), color),
        ]
        table.add_row(*cells)
    # Print the table to the console
    console.print(Panel("CVE SUMMARY", expand=False))
    console.print(table)

    cve_by_score = defaultdict(list)
    # group cve_data by its remarks
    for cve, cve_data in all_cve_data.items():
        key = (
            SUMMARY_ORDER[cve_data["severity"]],
            cve_data["severity"]
        )
        cve_by_score[key].append(
            {
                "vendor": cve_data["pkglist_package_vendor"],
                "product": cve_data["pkglist_package_name"],
                "version": cve_data["pkglist_package_version"],
                "cve_number": cve_data["cve_number"],
                "severity": cve_data["severity"],
                "score": cve_data["score"],
                "cvss_version": cve_data["cvss_version"],
                "version_range": cve_data["version_range"]
            }
        )

    for key in sorted(cve_by_score):
        score = key[1]
        color = score_colors(score)
        # table instance
        table = Table()

        # Add Head Columns to the Table
        table.add_column("Vendor")
        table.add_column("Product")
        table.add_column("Version")
        table.add_column("CVE Number")
        table.add_column("Severity")
        table.add_column("Score (CVSS Version)")
        table.add_column("Affected versions")
        # table.add_column("CVSS Version")

        for cve_data in cve_by_score[key]:
            color = cve_data["severity"].lower()
            cells = [
                Text.styled(cve_data["vendor"], color),
                Text.styled(cve_data["product"], color),
                Text.styled(cve_data["version"], color),
                linkify_cve(Text.styled(cve_data["cve_number"], color)),
                Text.styled(cve_data["severity"], color),
                Text.styled(
                    str(cve_data["score"])
                    + " (v"
                    + str(cve_data["cvss_version"])
                    + ")",
                    color,
                ),
                Text.styled(format_version_range(cve_data["version_range"]), color),
            ]
            table.add_row(*cells)
        # Print the table to the console
        console.print(table)
        for cve_data in cve_by_score[key]:
            if "*" in cve_data["vendor"]:
                console.print("* vendors guessed by the tool")
                break



import asyncio
import datetime
import glob
import gzip
import hashlib
import json
import logging
import os
import re
import shutil
import sqlite3

import aiohttp
from bs4 import BeautifulSoup
from rich.progress import track

from cve_ape.async_utils import FileIO, GzipFile, RateLimiter, run_coroutine
from cve_ape.error_handler import (
    AttemptedToWriteOutsideCachedir,
    CVEDataForCurlVersionNotInCache,
    CVEDataForYearNotInCache,
    ErrorHandler,
    ErrorMode,
    NVDRateLimit,
    SHAMismatch,
)
from cve_ape.log import LOGGER
from cve_ape.nvd_api import NVD_API
from cve_ape.version import check_latest_version

logging.basicConfig(level=logging.DEBUG)

# database defaults
DISK_LOCATION_DEFAULT = os.path.join(os.path.expanduser("~"), ".cache", "cve-ape")
DISK_LOCATION_BACKUP = os.path.join(
    os.path.expanduser("~"), ".cache", "cve-ape", "cache-backup"
)
DBNAME = "cve.db"
OLD_CACHE_DIR = os.path.join(os.path.expanduser("~"), ".cache", "cvedb")
NVD_FILENAME_TEMPLATE = "nvdcve-1.1-{}.json.gz"


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class CVEDB(metaclass=Singleton):
    """
    Downloads NVD data in json form and stores it on disk in a cache.
    """

    CACHEDIR = DISK_LOCATION_DEFAULT
    BACKUPCACHEDIR = DISK_LOCATION_BACKUP
    FEED = "https://nvd.nist.gov/vuln/data-feeds"
    LOGGER = LOGGER.getChild("CVEDB")
    NVDCVE_FILENAME_TEMPLATE = NVD_FILENAME_TEMPLATE
    CURL_CVE_FILENAME_TEMPLATE = "curlcve-{}.json"
    META_LINK = "https://nvd.nist.gov"
    META_REGEX = re.compile(r"\/feeds\/json\/.*-[0-9]*\.[0-9]*-[0-9]*\.meta")
    RANGE_UNSET = ""

    def __init__(
            self,
            feed=None,
            cachedir=None,
            backup_cachedir=None,
            version_check=True,
            session=None,
            error_mode=ErrorMode.TruncTrace,
            nvd_type="json",
            incremental_update=False,
            nvd_api_key: str = "",
    ):
        self.feed = feed if feed is not None else self.FEED
        self.cachedir = cachedir if cachedir is not None else self.CACHEDIR
        self.backup_cachedir = (
            backup_cachedir if backup_cachedir is not None else self.BACKUPCACHEDIR
        )
        self.error_mode = error_mode
        # Will be true if refresh was successful
        self.was_updated = False

        # version update
        self.version_check = version_check

        # set up the db if needed
        self.dbpath = os.path.join(self.cachedir, DBNAME)
        self.connection = None
        self.session = session
        self.cve_count = -1
        self.nvd_type = nvd_type
        self.incremental_update = incremental_update
        self.all_cve_entries = None

        # store the nvd api key for use later
        self.nvd_api_key = nvd_api_key

        if not os.path.exists(self.dbpath):
            self.rollback_cache_backup()

    def get_cve_count(self):
        if self.cve_count == -1:
            # Force update
            self.check_cve_entries()
        return self.cve_count

    async def nist_fetch_using_api(self):
        """Fetch using NVD's CVE API (as opposed to NVD's JSON Vulnerability Feeds)"""
        nvd_api = NVD_API(
            logger=self.LOGGER,
            error_mode=self.error_mode,
            incremental_update=self.incremental_update,
            api_key=self.nvd_api_key,
        )
        if self.incremental_update:
            await nvd_api.get_nvd_params(
                time_of_last_update=datetime.datetime.fromtimestamp(
                    self.get_db_update_date()
                )
            )
        else:
            await nvd_api.get_nvd_params()
        await nvd_api.get()
        await nvd_api.session.close()
        nvd_api.session = None
        return nvd_api.all_cve_entries

    def check_db_exists(self):
        return os.path.isfile(self.dbpath)

    def get_db_update_date(self):
        # last time when CVE data was updated
        self.time_of_last_update = datetime.datetime.fromtimestamp(
            os.path.getmtime(self.dbpath)
        )
        return os.path.getmtime(self.dbpath)

    async def getmeta(self, session, meta_url):
        async with await session.get(meta_url) as response:
            response.raise_for_status()
            return (
                meta_url.replace(".meta", ".json.gz"),
                dict(
                    [
                        line.split(":", maxsplit=1)
                        for line in (await response.text()).splitlines()
                        if ":" in line
                    ]
                ),
            )

    async def nist_scrape(self, session):
        async with await session.get(self.feed) as response:
            response.raise_for_status()
            page = await response.text()
            json_meta_links = self.META_REGEX.findall(page)
            return dict(
                await asyncio.gather(
                    *(
                        self.getmeta(session, f"{self.META_LINK}{meta_url}")
                        for meta_url in json_meta_links
                    )
                )
            )

    async def cache_update(self, session, url, sha, chunk_size=16 * 1024):
        """
        Update the cache for a single year of NVD data.
        """
        filename = url.split("/")[-1]
        # Ensure we only write to files within the cachedir
        filepath = os.path.abspath(os.path.join(self.cachedir, filename))
        if not filepath.startswith(os.path.abspath(self.cachedir)):
            with ErrorHandler(mode=self.error_mode, logger=self.LOGGER):
                raise AttemptedToWriteOutsideCachedir(filepath)
        # Validate the contents of the cached file
        if os.path.isfile(filepath):
            # Validate the sha and write out
            sha = sha.upper()
            calculate = hashlib.sha256()
            async with GzipFile(filepath, "rb") as f:
                chunk = await f.read(chunk_size)
                while chunk:
                    calculate.update(chunk)
                    chunk = await f.read(chunk_size)
            # Validate the sha and exit if it is correct, otherwise update
            gotsha = calculate.hexdigest().upper()
            if gotsha != sha:
                os.unlink(filepath)
                self.LOGGER.debug(
                    f"SHA mismatch for {filename} (have: {gotsha}, want: {sha})"
                )
            else:
                self.LOGGER.debug(f"Correct SHA for {filename}")
                return
        self.LOGGER.debug(f"Updating CVE cache for {filename}")
        async with await session.get(url) as response:
            # Raise better error message on ratelimit by NVD
            if response.status == 403:
                with ErrorHandler(mode=self.error_mode, logger=self.LOGGER):
                    raise NVDRateLimit(
                        f"{url} : download failed, you may have been rate limited."
                    )
            # Raise for all other 4xx errors
            response.raise_for_status()
            gzip_data = await response.read()
        json_data = gzip.decompress(gzip_data)
        gotsha = hashlib.sha256(json_data).hexdigest().upper()
        async with FileIO(filepath, "wb") as filepath_handle:
            await filepath_handle.write(gzip_data)
        # Raise error if there was an issue with the sha
        if gotsha != sha:
            # Remove the file if there was an issue
            # exit(100)
            os.unlink(filepath)
            with ErrorHandler(mode=self.error_mode, logger=self.LOGGER):
                raise SHAMismatch(f"{url} (have: {gotsha}, want: {sha})")

    @staticmethod
    async def get_curl_versions(session):
        regex = re.compile(r"vuln-(\d+.\d+.\d+)\.html")
        async with await session.get(
                "https://curl.haxx.se/docs/vulnerabilities.html"
        ) as response:
            response.raise_for_status()
            html = await response.text()
        matches = regex.finditer(html)
        return [match.group(1) for match in matches]

    async def download_curl_version(self, session, version):
        async with await session.get(
                f"https://curl.haxx.se/docs/vuln-{version}.html"
        ) as response:
            response.raise_for_status()
            html = await response.text()
        soup = BeautifulSoup(html, "html.parser")
        table = soup.find("table")
        if not table:
            return
        headers = table.find_all("th")
        headers = list(map(lambda x: x.text.strip().lower(), headers))
        self.LOGGER.debug(headers)
        rows = table.find_all("tr")
        json_data = []
        for row in rows:
            cols = row.find_all("td")
            values = (ele.text.strip() for ele in cols)
            data = dict(zip(headers, values))
            if data:
                json_data.append(data)
        filepath = os.path.abspath(
            os.path.join(self.cachedir, f"curlcve-{version}.json")
        )
        async with FileIO(filepath, "w") as f:
            await f.write(json.dumps(json_data, indent=4))

    async def refresh(self):
        """Refresh the cve database and check for new version."""
        # refresh the database
        if not os.path.isdir(self.cachedir):
            os.makedirs(self.cachedir)

        # check for the latest version
        if self.version_check:
            check_latest_version()
        if not self.session:
            connector = aiohttp.TCPConnector(limit_per_host=19)
            self.session = RateLimiter(
                aiohttp.ClientSession(connector=connector, trust_env=True)
            )

        tasks = []
        if self.nvd_type == "api":
            self.LOGGER.info("[Using NVD API]")
            self.all_cve_entries, curl_metadata = await asyncio.gather(
                self.nist_fetch_using_api(),
                self.get_curl_versions(self.session),
            )

        else:
            self.LOGGER.info("Downloading CVE data...")
            nvd_metadata, curl_metadata = await asyncio.gather(
                self.nist_scrape(self.session),
                self.get_curl_versions(self.session),
            )

            tasks = [
                self.cache_update(self.session, url, meta["sha256"])
                for url, meta in nvd_metadata.items()
                if meta is not None
            ]

        # We use gather to create a single task from a set of tasks
        # which download CVEs for each version of curl. Otherwise
        # the progress bar would show that we are closer to
        # completion than we think, because lots of curl CVEs (for
        # each version) have been downloaded
        tasks.append(
            asyncio.gather(
                *(
                    asyncio.ensure_future(
                        self.download_curl_version(self.session, version)
                    )
                    for version in curl_metadata
                )
            )
        )
        total_tasks = len(tasks)

        # error_mode.value will only be greater than 1 if quiet mode.
        if self.error_mode.value > 1:
            iter_tasks = track(
                asyncio.as_completed(tasks),
                description="Downloading CVEs...",
                total=total_tasks,
            )
        else:
            iter_tasks = asyncio.as_completed(tasks)

        for task in iter_tasks:
            await task
        self.was_updated = True
        await self.session.close()
        self.session = None

    def refresh_cache_and_update_db(self):
        self.LOGGER.debug("Updating CVE data. This will take a few minutes.")
        # refresh the nvd cache
        run_coroutine(self.refresh())

        # if the database isn't open, open it
        self.init_database()
        self.populate_db()

    def get_cvelist_if_stale(self):
        """Update if the local db is more than one day old.
        This avoids the full slow update with every execution.
        """
        if not os.path.isfile(self.dbpath) or (
                datetime.datetime.today()
                - datetime.datetime.fromtimestamp(os.path.getmtime(self.dbpath))
        ) > datetime.timedelta(hours=24):
            self.refresh_cache_and_update_db()
            self.time_of_last_update = datetime.datetime.today()
        else:
            self.time_of_last_update = datetime.datetime.fromtimestamp(
                os.path.getmtime(self.dbpath)
            )
            self.LOGGER.info(
                "Using cached CVE data (<24h old). Use -u now to update immediately."
            )
            self.db_open()
            if not self.latest_schema(self.connection.cursor()):
                self.refresh_cache_and_update_db()
                self.time_of_last_update = datetime.datetime.today()
            else:
                self.db_close()

    def latest_schema(self, cursor):
        """Check database is using latest schema"""
        self.LOGGER.debug("Check database is using latest schema")
        schema_check = "SELECT * FROM cve_severity WHERE 1=0"
        result = cursor.execute(schema_check)
        schema_latest = False
        # Look through column names and check for column added in latest schema
        for col_name in result.description:
            if col_name[0] == "cvss_vector":
                schema_latest = True
        return schema_latest

    def check_cve_entries(self):
        """Report if database has some CVE entries"""
        self.db_open()
        cursor = self.connection.cursor()
        cve_entries_check = "SELECT COUNT(*) FROM cve_severity"
        cursor.execute(cve_entries_check)
        # Find number of entries
        cve_entries = cursor.fetchone()[0]
        self.LOGGER.info(f"There are {cve_entries} CVE entries in the database")
        self.db_close()
        self.cve_count = cve_entries
        return cve_entries > 0

    def init_database(self):
        """Initialize db tables used for storing cve/version data"""
        self.db_open()
        cursor = self.connection.cursor()
        cve_data_create = """
        CREATE TABLE IF NOT EXISTS cve_severity (
            cve_number TEXT,
            severity TEXT,
            description TEXT,
            score INTEGER,
            cvss_version INTEGER,
            cvss_vector TEXT,
            PRIMARY KEY(cve_number)
        )
        """
        version_range_create = """
        CREATE TABLE IF NOT EXISTS cve_range (
            cve_number TEXT,
            vendor TEXT,
            product TEXT,
            version TEXT,
            versionStartIncluding TEXT,
            versionStartExcluding TEXT,
            versionEndIncluding TEXT,
            versionEndExcluding TEXT,
            FOREIGN KEY(cve_number) REFERENCES cve_severity(cve_number)
        )
        """

        index_range = "CREATE INDEX IF NOT EXISTS product_index ON cve_range (cve_number, vendor, product)"
        cursor.execute(cve_data_create)
        cursor.execute(version_range_create)
        cursor.execute(index_range)

        # Check that latest schema is being used
        if not self.latest_schema(cursor):
            # Recreate table using latest schema
            self.LOGGER.info("Upgrading database to latest schema")
            cursor.execute("DROP TABLE cve_severity")
            cursor.execute(cve_data_create)
            self.clear_cached_data()
        self.connection.commit()

    def populate_db(self):
        """Function that populates the database from the JSON.

        WARNING: After some inspection of the data, we are assuming that start/end ranges are kept together
        in single nodes.  This isn't *required* by the json so may not be true everywhere.  If that's the case,
        we'll need a better parser to match those together.
        """
        self.db_open()
        cursor = self.connection.cursor()

        insert_severity = """
        INSERT or REPLACE INTO cve_severity(
            CVE_number,
            severity,
            description,
            score,
            cvss_version,
            cvss_vector
        )
        VALUES (?, ?, ?, ?, ?, ?)
        """
        insert_cve_range = """
        INSERT or REPLACE INTO cve_range(
            cve_number,
            vendor,
            product,
            version,
            versionStartIncluding,
            versionStartExcluding,
            versionEndIncluding,
            versionEndExcluding
        ) VALUES (?, ?, ?, ?, ?, ?, ?, ?)
        """
        del_cve_range = "DELETE from cve_range where CVE_number=?"
        if self.nvd_type == "api":
            self.LOGGER.debug("Populating database from NVD API")
            self.populate_cve_list(
                self.all_cve_entries,
                cursor,
                insert_severity,
                insert_cve_range,
                del_cve_range,
            )
        else:
            # error_mode.value will only be greater than 1 if quiet mode.
            if self.error_mode.value > 1:
                years = track(self.nvd_years(), description="Updating CVEs from NVD...")
            else:
                years = self.nvd_years()

            for year in years:
                cve_data = self.load_nvd_year(year)
                self.LOGGER.debug(
                    f'Time = {datetime.datetime.today().strftime("%H:%M:%S")}'
                )
                self.populate_cve_list(
                    cve_data["CVE_Items"],
                    cursor,
                    insert_severity,
                    insert_cve_range,
                    del_cve_range,
                )
        self.connection.commit()

        # supplemental data gets added here
        self.supplement_curl()

        self.db_close()

    def populate_cve_list(
            self, all_cve_list, cursor, insert_severity, insert_cve_range, del_cve_range
    ):
        """Populate the cve_list with the CVEs from the JSON"""
        for cve_item in all_cve_list:
            # the information we want:
            # CVE ID, Severity, Score ->
            # affected {Vendor(s), Product(s), Version(s)}
            cve = {
                "ID": cve_item["cve"]["CVE_data_meta"]["ID"],
                "description": cve_item["cve"]["description"]["description_data"][0][
                    "value"
                ],
                "severity": "unknown",
                "score": "unknown",
                "CVSS_version": "unknown",
                "CVSS_vector": "unknown",
            }
            if cve["description"].startswith("** REJECT **"):
                # Skip this CVE if it's marked as 'REJECT'
                continue

            # Get CVSSv3 or CVSSv2 score for output.
            # Details are left as an exercise to the user.
            if "baseMetricV3" in cve_item["impact"]:
                cve["severity"] = cve_item["impact"]["baseMetricV3"]["cvssV3"][
                    "baseSeverity"
                ]
                cve["score"] = cve_item["impact"]["baseMetricV3"]["cvssV3"]["baseScore"]
                cve["CVSS_vector"] = cve_item["impact"]["baseMetricV3"]["cvssV3"][
                    "vectorString"
                ]
                cve["CVSS_version"] = 3
            elif "baseMetricV2" in cve_item["impact"]:
                cve["severity"] = cve_item["impact"]["baseMetricV2"]["severity"]
                cve["score"] = cve_item["impact"]["baseMetricV2"]["cvssV2"]["baseScore"]
                cve["CVSS_vector"] = cve_item["impact"]["baseMetricV2"]["cvssV2"][
                    "vectorString"
                ]
                cve["CVSS_version"] = 2

            # self.LOGGER.debug(
            #    "Severity: {} ({}) v{}".format(
            #        CVE["severity"], CVE["score"], CVE["CVSS_version"]
            #    )
            # )

            cursor.execute(
                insert_severity,
                [
                    cve["ID"],
                    cve["severity"],
                    cve["description"],
                    cve["score"],
                    cve["CVSS_version"],
                    cve["CVSS_vector"],
                ],
            )

            # Delete any old range entries for this CVE_number
            cursor.execute(del_cve_range, (cve["ID"],))

            # walk the nodes with version data
            # return list of versions
            affects_list = []
            if "configurations" in cve_item:
                for node in cve_item["configurations"]["nodes"]:
                    # self.LOGGER.debug("NODE: {}".format(node))
                    affects_list.extend(self.parse_node(node))
                    if "children" in node:
                        for child in node["children"]:
                            affects_list.extend(self.parse_node(child))
            # self.LOGGER.debug("Affects: {}".format(affects_list))
            cursor.executemany(
                insert_cve_range,
                [
                    (
                        cve["ID"],
                        affected["vendor"],
                        affected["product"],
                        affected["version"],
                        affected["versionStartIncluding"],
                        affected["versionStartExcluding"],
                        affected["versionEndIncluding"],
                        affected["versionEndExcluding"],
                    )
                    for affected in affects_list
                ],
            )

    def parse_node(self, node):
        affects_list = []
        if "cpe_match" in node:
            for cpe_match in node["cpe_match"]:
                # self.LOGGER.debug(cpe_match["cpe23Uri"])
                cpe_split = cpe_match["cpe23Uri"].split(":")
                affects = {
                    "vendor": cpe_split[3],
                    "product": cpe_split[4],
                    "version": cpe_split[5],
                }
                range_fields = [
                    "versionStartIncluding",
                    "versionStartExcluding",
                    "versionEndIncluding",
                    "versionEndExcluding",
                ]
                for field in range_fields:
                    if field in cpe_match:
                        affects[field] = cpe_match[field]
                    else:
                        affects[field] = self.RANGE_UNSET

                affects_list.append(affects)
        return affects_list

    def supplement_curl(self):
        """
        Get additional CVE data directly from the curl website amd add it to the cvedb
        """
        self.db_open()
        insert_cve_range = """
        INSERT or REPLACE INTO cve_range(
            cve_number,
            vendor,
            product,
            version,
            versionStartIncluding,
            versionStartExcluding,
            versionEndIncluding,
            versionEndExcluding
        ) VALUES (?, ?, ?, ?, ?, ?, ?, ?)
        """
        cursor = self.connection.cursor()
        # No need to track this. It is very fast!
        for version in self.curl_versions():
            cve_list = self.load_curl_version(version)
            # for cve in cve_list:
            cursor.executemany(
                insert_cve_range,
                [
                    (
                        cve["cve"],
                        "haxx",
                        "curl",
                        version,
                        cve["from version"],
                        "",
                        cve["to and including"],
                        "",
                    )
                    for cve in cve_list
                ],
            )
            self.connection.commit()

    def load_nvd_year(self, year):
        """
        Return the dict of CVE data for the given year.
        """
        filename = os.path.join(
            self.cachedir, self.NVDCVE_FILENAME_TEMPLATE.format(year)
        )
        # Check if file exists
        if not os.path.isfile(filename):
            with ErrorHandler(mode=self.error_mode, logger=self.LOGGER):
                raise CVEDataForYearNotInCache(year)
        # Open the file and load the JSON data, log the number of CVEs loaded
        with gzip.open(filename, "rb") as fileobj:
            cves_for_year = json.load(fileobj)
            self.LOGGER.debug(
                f'Year {year} has {len(cves_for_year["CVE_Items"])} CVEs in dataset'
            )
            return cves_for_year

    def nvd_years(self):
        """
        Return the years we have NVD data for.
        """
        return sorted(
            int(filename.split(".")[-3].split("-")[-1])
            for filename in glob.glob(
                os.path.join(self.cachedir, "nvdcve-1.1-*.json.gz")
            )
        )

    def load_curl_version(self, version):
        """
        Return the dict of CVE data for the given curl version.
        """
        filename = os.path.join(
            self.cachedir, self.CURL_CVE_FILENAME_TEMPLATE.format(version)
        )
        # Check if file exists
        if not os.path.isfile(filename):
            with ErrorHandler(mode=self.error_mode, logger=self.LOGGER):
                raise CVEDataForCurlVersionNotInCache(version)
        # Open the file and load the JSON data, log the number of CVEs loaded
        with open(filename, "rb") as fileobj:
            cves_for_version = json.load(fileobj)
            self.LOGGER.debug(
                f"Curl Version {version} has {len(cves_for_version)} CVEs in dataset"
            )
            return cves_for_version

    def curl_versions(self):
        """
        Return the versions we have Curl data for.
        """
        regex = re.compile(r"curlcve-(\d+.\d+.\d).json")
        return [
            regex.search(filename).group(1)
            for filename in glob.glob(os.path.join(self.cachedir, "curlcve-*.json"))
        ]

    def clear_cached_data(self):
        self.create_cache_backup()
        if os.path.exists(self.cachedir):
            self.LOGGER.warning(f"Updating cachedir {self.cachedir}")
            shutil.rmtree(self.cachedir)
        # Remove files associated with pre-1.0 development tree
        if os.path.exists(OLD_CACHE_DIR):
            self.LOGGER.warning(f"Deleting old cachedir {OLD_CACHE_DIR}")
            shutil.rmtree(OLD_CACHE_DIR)

    def get_cve_info_bulk(self, cve_list):
        self.db_open()
        cursor = self.connection.cursor()
        query_temptbl_create = """
        CREATE TEMP TABLE IF NOT EXISTS project_cve_list (
        cve_number TEXT
        )
        """
        query_temptbl_insert = """
        INSERT or REPLACE INTO project_cve_list (cve_number) VALUES (?)
        """

        cursor.execute(query_temptbl_create)
        if type(cve_list) is not list or len(cve_list) == 0:
            return {}
        cves_in_tuple = tuple(map(lambda x: (x,), cve_list))
        cursor.executemany(query_temptbl_insert, cves_in_tuple)
        # self.connection.commit()

        query = """
        SELECT DISTINCT cve_severity.cve_number, severity, description, score, cvss_version, cvss_vector
                    FROM cve_severity INNER JOIN project_cve_list
            WHERE cve_severity.cve_number = project_cve_list.cve_number
        """
        cursor.execute(query)
        columnNames = [column[0] for column in cursor.description]
        cve_detailed_list = {}
        for row in cursor.fetchall():
            cve_detailed_list[row[0]] = dict(zip(columnNames, row))

        self.db_close()
        return cve_detailed_list

    def get_cves_for_packages(self, pkglist_package_names):
        self.db_open()
        cursor = self.connection.cursor()
        query_temptbl_create = """
        CREATE TEMP TABLE IF NOT EXISTS project_cve_packages (
        package TEXT,
        version TEXT
        )
        """
        query_temptbl_insert = """
        INSERT or REPLACE INTO project_cve_packages (package, version) VALUES (?, ?)
        """

        cursor.execute(query_temptbl_create)
        pkglist_packages_in_tuple = tuple(map(lambda x: (x["name"], x["version"]), pkglist_package_names))
        cursor.executemany(query_temptbl_insert, pkglist_packages_in_tuple)
        self.connection.commit()

        query = """
        SELECT cve_range.CVE_number,
            cve_range.versionStartIncluding,
            cve_range.versionStartExcluding,
            cve_range.versionEndIncluding,
            cve_range.versionEndExcluding,
            cve_range.version,
            project_cve_packages.package,
            project_cve_packages.version as pkglist_package_version,
            cve_range.vendor as vendor
            FROM project_cve_packages INNER JOIN cve_range
            WHERE project_cve_packages.package = cve_range.product
        """
        cursor.execute(query)
        columnNames = [column[0] for column in cursor.description]
        cve_list = list(map(lambda x: dict(zip(columnNames, x)), cursor.fetchall()))

        self.db_close()
        return cve_list

    def get_packages_from_pkglist(self, pkglist_package_names):
        self.db_open()
        cursor = self.connection.cursor()
        query_temptbl_create = """
        CREATE TEMP TABLE IF NOT EXISTS project_packages (
        package TEXT,
        version TEXT
        )
        """
        query_temptbl_insert = """
        INSERT or REPLACE INTO project_packages (package, version) VALUES (?, ?)
        """

        cursor.execute(query_temptbl_create)
        pkglist_packages_in_tuple = tuple(map(lambda x: (x["name"], x["version"]), pkglist_package_names))
        cursor.executemany(query_temptbl_insert, pkglist_packages_in_tuple)
        self.connection.commit()

        query = """
        SELECT DISTINCT cve_range.product, cve_range.vendor FROM project_packages INNER JOIN cve_range
        WHERE project_packages.package = cve_range.product
        """
        cursor.execute(query)
        known_packages = []
        pkglist_packages_in_dict = dict(pkglist_packages_in_tuple)
        for x in cursor.fetchall():
            data = {"name": x[0], "vendor": x[1], "version": pkglist_packages_in_dict[x[0]]}
            known_packages.append(data)
        known_packages_scoped_names = list(map(lambda x: x["name"], known_packages))

        unknowns_counter = 0
        unknown_packages = []
        for package_name in pkglist_package_names:
            if package_name["name"] not in known_packages_scoped_names:
                unknowns_counter += 1
                unknown_packages.append(package_name)

        self.LOGGER.warning("Unknown packages: {} out of {}".format(unknowns_counter, len(pkglist_package_names)))

        self.db_close()
        return known_packages, unknown_packages

    def db_open(self):
        """Opens connection to sqlite database."""
        if not self.connection:
            self.connection = sqlite3.connect(self.dbpath)

    def db_close(self):
        """Closes connection to sqlite database."""
        if self.connection:
            self.connection.close()
            self.connection = None

    def create_cache_backup(self):
        """Creates a backup of the cachedir in case anything fails"""
        if os.path.exists(self.cachedir):
            self.LOGGER.debug(
                f"Creating backup of cachedir {self.cachedir} at {self.backup_cachedir}"
            )
            self.remove_cache_backup()
            shutil.copytree(self.cachedir, self.backup_cachedir)

    def remove_cache_backup(self):
        """Removes the backup if database was successfully loaded"""
        if os.path.exists(self.backup_cachedir):
            self.LOGGER.debug(f"Removing backup cache from {self.backup_cachedir}")
            shutil.rmtree(self.backup_cachedir)

    def rollback_cache_backup(self):
        """Rollback the cachedir backup in case anything fails"""
        if os.path.exists(os.path.join(self.backup_cachedir, DBNAME)):
            self.LOGGER.info("Rolling back the cache to its previous state")
            if os.path.exists(self.cachedir):
                shutil.rmtree(self.cachedir)
            shutil.move(self.backup_cachedir, self.cachedir)

    def __del__(self):
        self.rollback_cache_backup()



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

""" Utility classes for the CVE Binary Tool """
from __future__ import annotations

import fnmatch
import os
import sys
from enum import Enum
from typing import DefaultDict, Iterator, List, NamedTuple, Pattern, Set, Union


class OrderedEnum(Enum):
    def __ge__(self, other: OrderedEnum) -> bool:
        if self.__class__ is other.__class__:
            return self.value >= other.value
        return NotImplemented

    def __gt__(self, other: OrderedEnum) -> bool:
        if self.__class__ is other.__class__:
            return self.value > other.value
        return NotImplemented

    def __le__(self, other: OrderedEnum) -> bool:
        if self.__class__ is other.__class__:
            return self.value <= other.value
        return NotImplemented

    def __lt__(self, other: OrderedEnum) -> bool:
        if self.__class__ is other.__class__:
            return self.value < other.value
        return NotImplemented


class Remarks(OrderedEnum):
    NewFound = 1, "1", "NewFound", "n", "N"
    Unexplored = 2, "2", "Unexplored", "u", "U", ""
    Confirmed = 3, "3", "Confirmed", "c", "C"
    Mitigated = 4, "4", "Mitigated", "m", "M"
    Ignored = 5, "5", "Ignored", "i", "I"

    def __new__(cls, value: int, *aliases: str) -> Remarks:
        obj = object.__new__(cls)
        obj._value_ = value
        for alias in aliases:
            cls._value2member_map_[alias] = obj
        return obj


class CVE(NamedTuple):
    cve_number: str
    severity: str
    remarks: Remarks = Remarks.NewFound
    description: str = ""
    comments: str = ""
    score: float = 0
    cvss_version: int = 0
    cvss_vector: str = ""


class ProductInfo(NamedTuple):
    vendor: str
    product: str
    version: str


class ScanInfo(NamedTuple):
    product_info: ProductInfo
    file_path: str


class VersionInfo(NamedTuple):
    start_including: str
    start_excluding: str
    end_including: str
    end_excluding: str
    is_conditional: bool


class CVEData(DefaultDict[str, Union[List[CVE], Set[str]]]):
    def __missing__(self, key: str) -> list[CVE] | set[str]:
        if key == "cves":
            new_list: list[CVE] = []
            self[key] = new_list
        elif key == "paths":
            new_set: set[str] = set()
            self[key] = new_set
        else:
            return NotImplemented
        return self[key]


def regex_find(lines: str, version_patterns: list[Pattern[str]]) -> str:
    """Search a set of lines to find a match for the given regex"""
    new_guess = ""

    for pattern in version_patterns:
        match = pattern.search(lines)
        if match:
            new_guess = match.group(1).strip()
            break
    if new_guess != "":
        new_guess = new_guess.replace("_", ".")
        return new_guess.replace("-", ".")
    else:
        return "UNKNOWN"


def inpath(binary: str) -> bool:
    """Check to see if something is available in the path.
    Used to check if dependencies are installed before use."""
    if sys.platform == "win32":
        return any(
            list(
                map(
                    lambda dirname: os.path.isfile(
                        os.path.join(dirname, binary + ".exe")
                    ),
                    os.environ.get("PATH", "").split(";"),
                )
            )
        )
    return any(
        list(
            map(
                lambda dirname: os.path.isfile(os.path.join(dirname, binary)),
                os.environ.get("PATH", "").split(":"),
            )
        )
    )


class DirWalk:
    """
    for filename in DirWalk('*.c').walk(roots):
        do a thing with the c-files in the roots directories
    """

    def __init__(
        self,
        pattern: str = "*",
        folder_include_pattern: str = "*",
        folder_exclude_pattern: str = ".git",
        file_exclude_pattern: str = "",
        yield_files: bool = True,
        yield_folders: bool = False,
    ) -> None:
        """
        Generator for walking the file system and filtering the results.
        """
        self.pattern = pattern
        self.folder_include_pattern = folder_include_pattern
        self.folder_exclude_pattern = folder_exclude_pattern
        self.file_exclude_pattern = file_exclude_pattern
        self.yield_files = yield_files
        self.yield_folders = yield_folders

    def walk(self, roots: list[str] | None = None) -> Iterator[str]:
        """Walk the directory looking for files matching the pattern"""
        if roots is None:
            roots = []
        for root in roots:
            for dirpath, dirnames, filenames in os.walk(root):
                # Filters
                filenames[:] = [
                    filename
                    for filename in filenames
                    if self.pattern_match(os.path.join(dirpath, filename), self.pattern)
                    and not self.pattern_match(
                        os.path.join(dirpath, filename), self.file_exclude_pattern
                    )
                    and not self.pattern_match(
                        os.path.join(dirpath, filename), self.folder_exclude_pattern
                    )
                ]
                dirnames[:] = [
                    dirname
                    for dirname in dirnames
                    if self.pattern_match(
                        os.path.join(dirpath, dirname), self.folder_include_pattern
                    )
                    and not self.pattern_match(
                        os.path.join(dirpath, dirname), self.folder_exclude_pattern
                    )
                ]
                # Yields
                if self.yield_files:
                    for filename in filenames:
                        yield os.path.normpath(os.path.join(dirpath, filename))
                if self.yield_folders:
                    for dirname in dirnames:
                        yield os.path.normpath(os.path.join(dirpath, dirname))

    @staticmethod
    def pattern_match(text: str, patterns: str) -> bool:
        """Match filename patterns"""
        if not patterns:
            return False
        for pattern in patterns.split(";"):
            if fnmatch.fnmatch(text, pattern):
                return True
        return False



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

import csv
import json
import os
from logging import Logger
from typing import IO, Dict

from .console import output_console
from .util import (
    add_extension_if_not,
    format_output,
    generate_filename,
    get_cve_summary,
)
from ..cve_scanner import CVEData
from ..error_handler import ErrorHandler, ErrorMode
from ..log import LOGGER
from ..util import ProductInfo


def output_json(all_cve_data: Dict[ProductInfo, CVEData], outfile: IO):
    """Output a JSON of CVEs"""
    formatted_output = format_output(all_cve_data)
    json.dump(formatted_output, outfile, indent="    ")


def output_csv(all_cve_data: Dict[ProductInfo, CVEData], outfile):
    """Output a CSV of CVEs"""
    formatted_output = format_output(all_cve_data)
    writer = csv.DictWriter(
        outfile,
        fieldnames=[
            "vendor",
            "product",
            "version",
            "cve_number",
            "severity",
            "score",
            "cvss_version",
            "cvss_vector",
            "paths",
            "remarks",
            "comments",
        ],
    )
    writer.writeheader()
    writer.writerows(formatted_output)


class OutputEngine:
    def __init__(
        self,
        all_cve_data: {},
        filename: str,
        logger: Logger = None,
    ):
        self.logger = logger or LOGGER.getChild(self.__class__.__name__)
        self.filename = os.path.abspath(filename) if filename else ""
        self.all_cve_data = all_cve_data

    def output_cves(self, outfile, output_type="console"):
        if output_type == "json":
            output_json(self.all_cve_data, outfile)
        elif output_type == "csv":
            output_csv(self.all_cve_data, outfile)
        else:  # console, or anything else that is unrecognised
            output_console(
                self.all_cve_data
            )

    def output_file(self, output_type="console"):

        """Generate a file for list of CVE"""

        if output_type == "console":
            # short circuit file opening logic if we are actually
            # just writing to stdout
            self.output_cves(self.filename, output_type)
            return

        # Check if we need to generate a filename
        if not self.filename:
            self.filename = generate_filename(output_type)
        else:
            # check and add if the filename doesn't contain extension
            self.filename = add_extension_if_not(self.filename, output_type)

            self.filename = self.check_file_path(self.filename, output_type)

            # try opening that file
            with ErrorHandler(mode=ErrorMode.Ignore) as e:
                with open(self.filename, "w") as f:
                    f.write("testing")
                os.remove(self.filename)
            if e.exit_code:
                self.logger.info(
                    f"Exception {e.exc_val} occurred while writing to the file {self.filename} "
                    "Switching Back to Default Naming Convention"
                )
                self.filename = generate_filename(output_type)

        # call to output_cves
        mode = "w"
        with open(self.filename, mode) as f:
            self.output_cves(f, output_type)

    def check_file_path(self, filepath: str, output_type: str, prefix: str = "output"):
        # check if the file already exists
        if os.path.isfile(filepath):
            self.logger.warning(f"Failed to write at '{filepath}'. File already exists")
            self.logger.info("Generating a new filename with Default Naming Convention")
            filepath = generate_filename(output_type, prefix)

        return filepath

    def check_dir_path(
        self, filepath: str, output_type: str, prefix: str = "intermediate"
    ):

        if os.path.isdir(filepath):
            self.logger.info(
                f"Generating a new filename with Default Naming Convention in directory path {filepath}"
            )
            filename = os.path.basename(generate_filename(output_type, prefix))
            filepath = os.path.join(filepath, filename)

        return filepath



import os
from collections import defaultdict
from datetime import datetime
from typing import DefaultDict, Dict, List

from ..util import CVE, Remarks


def get_cve_summary(all_cve_data):
    summary = {"CRITICAL": 0, "HIGH": 0, "MEDIUM": 0, "LOW": 0}

    for cve_data in all_cve_data.values():
        for s in summary.keys():
            summary[s] += 1 if cve_data["severity"] == s else 0
    return summary


def generate_filename(extension: str, prefix: str = "output") -> str:
    now = datetime.now().strftime("%Y-%m-%d.%H-%M-%S")

    filename = os.path.abspath(
        os.path.join(os.getcwd(), f"{prefix}.cve-bin-tool.{now}.{extension}")
    )

    return filename


def format_output(all_cve_data: []) -> List[Dict[str, str]]:
    formatted_output = []
    for product_info, cve_data in all_cve_data.items():
        for cve in cve_data["cves"]:
            formatted_output.append(
                {
                    "vendor": product_info.vendor,
                    "product": product_info.product,
                    "version": product_info.version,
                    "cve_number": cve.cve_number,
                    "severity": cve.severity,
                    "score": str(cve.score),
                    "cvss_version": str(cve.cvss_version),
                    "cvss_vector": cve.cvss_vector,
                    "paths": ", ".join(cve_data["paths"]),
                    "remarks": cve.remarks.name,
                    "comments": cve.comments,
                }
            )

    return formatted_output


def add_extension_if_not(filename: str, output_type: str) -> str:
    if not filename.endswith(f".{output_type}"):
        updated_filename = f"{filename}.{output_type}"
        return updated_filename
    else:
        return filename


def group_cve_by_remark(
    cve_by_product: List[CVE],
) -> DefaultDict[Remarks, List[Dict[str, str]]]:
    cve_by_remarks: DefaultDict[Remarks, List[Dict[str, str]]] = defaultdict(list)
    for cve in cve_by_product:
        cve_by_remarks[cve.remarks].append(
            {
                "cve_number": cve.cve_number,
                "severity": cve.severity,
                "description": cve.description,
                "vector": cve.cvss_vector,
            }
        )
    return cve_by_remarks



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

"""Logging"""
import logging

from rich.logging import RichHandler


# A log filter to filter out logs based on filter level
# Any log above and equal the specified level will not be logged
class LevelFilter(logging.Filter):
    def __init__(self, level):
        super().__init__()
        self.level = level

    def filter(self, record):
        return record.levelno < self.level


# Rich Handler by default Initalize a Console with stderr stream for logs
logging.basicConfig(
    level="INFO",
    format="%(name)s - %(message)s",
    datefmt="[%X]",
    handlers=[RichHandler()],
)

# Add the handlers to the root logger
root_logger = logging.getLogger()

LOGGER = logging.getLogger(__package__)
LOGGER.setLevel(logging.INFO)



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

import sys
from enum import Enum
from logging import Logger
from typing import Union

from rich.console import Console
from rich.traceback import Traceback

CONSOLE = Console(file=sys.stderr)


class InsufficientArgs(Exception):
    """Insufficient command line arguments"""


class EmptyTxtError(Exception):
    """Given txt File is empty"""


class InvalidListError(Exception):
    """Given File is an invalid package list"""


class InvalidCsvError(Exception):
    """Given File is an Invalid CSV"""


class InvalidCheckerError(Exception):
    """Raised when data provided to Checker is not correct"""


class MissingFieldsError(Exception):
    """Missing needed fields"""


class InvalidJsonError(Exception):
    """Given File is an Invalid JSON"""


class InvalidIntermediateJsonError(Exception):
    """Given Intermediate File is not in valid Format"""


class EmptyCache(Exception):
    """
    Raised when NVD is opened when verify=False and there are no files in the
    cache.
    """


class CVEDataForYearNotInCache(Exception):
    """
    Raised when the CVE data for a year is not present in the cache.
    """


class CVEDataForCurlVersionNotInCache(Exception):
    """
    Raised when the CVE data for a curl version is not present in the cache.
    """


class CVEDataMissing(Exception):
    """
    Raised when no CVE data is present in the database.
    """


class AttemptedToWriteOutsideCachedir(Exception):
    """
    Raised if we attempted to write to a file that would have been outside the
    cachedir.
    """


class NVDRateLimit(Exception):
    """
    Raised if you have been ratelimited by NVD.
    """


class NVDServiceError(Exception):
    """
    Raised if the NVD API returns a 503 error.
    """


class NVDKeyError(Exception):
    """
    Raised if the NVD API key is invalid.
    """


class SHAMismatch(Exception):
    """
    Raised if the sha of a file in the cache was not what it should be.
    """


class ExtractionFailed(ValueError):
    """Extraction fail"""


class UnknownArchiveType(ValueError):
    """Unknown archive type"""


class UnknownConfigType(Exception):
    """Unknown configuration file type"""


class ErrorMode(Enum):
    Ignore = 0
    NoTrace = 1
    TruncTrace = 2
    FullTrace = 3


def excepthook(exc_type, exc_val, exc_tb):
    trace = Traceback.from_exception(exc_type, exc_val, exc_tb)
    CONSOLE.print(trace)
    if ERROR_CODES.get(exc_type):
        sys.exit(ERROR_CODES[exc_type])


class ErrorHandler:
    """Error handler context manager.

    Supports Different modes like ignore error, print full trace, truncated trace and no trace.
    Log messages if logger specified.

    Args:
        mode (ErrorMode): Can take any valid ErrorMode as an arg and change output according to that.
        logger (Logger): logs error message specified while raising Exception if logger is passed
                         while class initialization.
                         Ex: raise ValueError("file required") will log 'ValueError: file required'.
    Attributes:
        exit_code (int): Stores exit code for raised Exception, set -1 for generic Exception and 0
                         if code executes without Exception.
        exc_val (Union[BaseException, None]): Stores exception instance if Exception raises otherwise None.
    """

    exit_code: int
    exc_val: Union[BaseException, None]

    def __init__(self, logger: Logger = None, mode: ErrorMode = ErrorMode.TruncTrace):
        self.mode = mode
        self.logger = logger
        self.exit_code = 0
        self.exc_val = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if isinstance(exc_val, BaseException):
            self.exit_code = ERROR_CODES.get(exc_type, -1)
            self.exc_val = exc_val
        if self.mode == ErrorMode.Ignore:
            return True
        if exc_type:
            if self.logger and exc_val:
                self.logger.error(f"{exc_type.__name__}: {exc_val}")
            if self.mode == ErrorMode.NoTrace:
                sys.exit(self.exit_code)
            if self.mode == ErrorMode.TruncTrace:
                CONSOLE.print_exception()
                sys.exit(self.exit_code)
            return False


# Exit codes for Exception.
# Error code 1 is reserved to mean that cves were found
# Error code 2 is reserved for argparse errors (default argparse behaviour)
# Error code 3 is reserved for "we found negative cves" (should be impossible)
# Error code 4-20 are reserved just in case
ERROR_CODES = {
    SystemExit: 2,
    FileNotFoundError: 21,
    InvalidCsvError: 22,
    InvalidJsonError: 22,
    EmptyTxtError: 22,
    InvalidListError: 22,
    MissingFieldsError: 23,
    InsufficientArgs: 24,
    EmptyCache: 25,
    CVEDataForYearNotInCache: 26,
    CVEDataForCurlVersionNotInCache: 27,
    AttemptedToWriteOutsideCachedir: 28,
    SHAMismatch: 29,
    ExtractionFailed: 30,
    UnknownArchiveType: 31,
    UnknownConfigType: 32,
    CVEDataMissing: 33,
    InvalidCheckerError: 34,
    NVDRateLimit: 35,
    InvalidIntermediateJsonError: 36,
    NVDServiceError: 37,
}



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: GPL-3.0-or-later

from rich.theme import Theme

# Rich theme to colorize in the terminal
cve_theme = Theme(
    {
        "critical": "red",
        "high": "blue",
        "medium": "yellow",
        "low": "green",
        "unknown": "white",
    }
)



import ast
import os
import sys
from io import StringIO

from setuptools.dist import Distribution

with open(os.path.join("cve_ape", "version.py")) as f:
    for line in f:
        if line.startswith("VERSION"):
            VERSION = ast.literal_eval(line.strip().split("=")[-1].strip())
            break


def IS_DEVELOP() -> bool:
    return any(
        list(
            map(
                os.path.isfile,
                list(
                    map(
                        lambda syspath: os.path.join(syspath, "cve-ape.egg-link"),
                        sys.path,
                    )
                ),
            )
        )
    )


def update_egg() -> None:
    with StringIO() as f:
        cwd = os.getcwd()
        os.chdir(os.path.join(os.path.dirname(__file__), ".."))
        sys.stdout = f
        sys.stderr = f
        dist = Distribution(
            dict(
                script_name="setup.py",
                script_args=["egg_info"],
                name="cve-ape",
                version=VERSION,
                entry_points={
                    "console_scripts": [
                        "cve-bin-tool = cve_bin_tool.cli:main",
                    ],
                },
            )
        )
        dist.parse_command_line()
        dist.run_commands()
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__
        os.chdir(cwd)


if __name__ == "__main__":
    update_egg()



import os
import re
import sqlite3
import sys
from collections import defaultdict
from logging import Logger
from string import ascii_lowercase
from typing import DefaultDict, Dict

from packaging.version import Version, LegacyVersion, InvalidVersion
from rich.console import Console

from cve_ape.cvedb import DBNAME, DISK_LOCATION_DEFAULT, CVEDB
from cve_ape.error_handler import ErrorMode
from cve_ape.log import LOGGER
from cve_ape.theme import cve_theme
from cve_ape.util import CVEData, ProductInfo, VersionInfo

ALPHA_TO_NUM: Dict[str, int] = dict(zip(ascii_lowercase, range(26)))
re_openssllike = re.compile(r'^((\d+(\.|)){3,})(.*)$')
re_digit = re.compile(r'\d+')


def break_openssllike_version(version):
    v = re.search(re_openssllike, version)
    if v:
        version = v.group(1)
        tail = v.group(4)
        dk = re.findall(re_digit, tail)
        if tail.isalpha():
            version = "{}.{}".format(version, ALPHA_TO_NUM[tail])
        elif dk:
            for d in dk:
                version = "{}.{}".format(version, d)
    return Version(version)


def ver(version):
    try:
        v = Version(version)
    except InvalidVersion:
        v = LegacyVersion(version)
    try:
        if re.search('[a-zA-Z]', v.base_version):
            v = break_openssllike_version(version)
    except:
        return Version("0")
    return v


def version_range(version, start_including, start_excluding, end_including, end_excluding):
    if start_including and end_including:
        return True if start_including <= version <= end_including else False
    if start_including and end_excluding:
        return True if start_including <= version < end_excluding else False
    if start_excluding and end_including:
        return True if start_excluding < version <= end_including else False
    if start_excluding and end_excluding:
        return True if start_excluding < version < end_excluding else False
    if start_including:
        return True if start_including <= version else False
    if start_excluding:
        return True if start_excluding < version else False
    if end_including:
        return True if end_including >= version else False
    if end_excluding:
        return True if end_excluding > version else False
    return False


class CVEScanner:
    """
    This class is for reading CVEs from the database
    """

    products_with_cve: int
    products_without_cve: int
    all_cve_data: DefaultDict[ProductInfo, CVEData]
    all_cve_version_info: Dict[str, VersionInfo]

    RANGE_UNSET: str = ""
    dbname: str = os.path.join(DISK_LOCATION_DEFAULT, DBNAME)
    CONSOLE: Console = Console(file=sys.stderr, theme=cve_theme)
    ALPHA_TO_NUM: Dict[str, int] = dict(zip(ascii_lowercase, range(26)))

    def __init__(
            self,
            score: int = 0,
            logger: Logger = None,
            error_mode: ErrorMode = ErrorMode.TruncTrace,
    ):
        self.logger = logger or LOGGER.getChild(self.__class__.__name__)
        self.error_mode = error_mode
        self.score = score
        self.products_with_cve = 0
        self.products_without_cve = 0
        self.all_cve_data = defaultdict(CVEData)
        self.all_cve_version_info = dict()

    def get_cves(self, packages):
        cve_db = CVEDB()
        cve_list = cve_db.get_cves_for_packages(packages)
        cve_report_list = {}

        for cve_bind_product in cve_list:
            cve_number = cve_bind_product["cve_number"]
            version_start_including = None if cve_bind_product["versionStartIncluding"] == "" \
                else ver(cve_bind_product["versionStartIncluding"])

            version_end_including = None if cve_bind_product["versionEndIncluding"] == "" \
                else ver(cve_bind_product["versionEndIncluding"])

            version_start_excluding = None if cve_bind_product["versionStartExcluding"] == "" \
                else ver(cve_bind_product["versionStartExcluding"])

            version_end_excluding = None if cve_bind_product["versionEndExcluding"] == "" \
                else ver(cve_bind_product["versionEndExcluding"])

            version = "*" if cve_bind_product["version"] == "*" else ver(cve_bind_product["version"])

            package_version = ver(cve_bind_product["pkglist_package_version"])

            passed = False
            conditional = False

            if version == "*":
                # including
                passed = version_range(package_version,
                                       version_start_including,
                                       version_start_excluding,
                                       version_end_including,
                                       version_end_excluding)
            elif version == package_version:
                passed = True
            elif version.base_version == package_version.base_version:
                passed = True  # conditional true
                conditional = True

            if passed:
                v = VersionInfo(
                    version_start_including,
                    version_start_excluding,
                    version_end_including,
                    version_end_excluding,
                    conditional)
                cve_report_list[cve_number] = {"pkglist_package_version": package_version.public,
                                               "pkglist_package_name": cve_bind_product["package"],
                                               "pkglist_package_vendor": cve_bind_product["vendor"],
                                               "version_range": v}

        # Go through and get all the severities
        details = cve_db.get_cve_info_bulk(list(cve_report_list.keys()))
        for k in cve_report_list.keys():
            try:
                cve_report_list[k] = {**cve_report_list[k], **details[k]}
            except:
                pass
        if cve_report_list.keys():
            for k in cve_report_list.keys():
                c = cve_report_list[k]
                self.logger.debug(
                    ' CVE in {}.{} v{} - {}:{}'.format(c["pkglist_package_name"],
                                                       c["pkglist_package_vendor"],
                                                       c["pkglist_package_version"],
                                                       k,
                                                       c["severity"])
                )
        return cve_report_list

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



import textwrap

import requests
from packaging import version

from cve_ape.log import LOGGER

VERSION: str = "1.0.4"


def check_latest_version():
    """Checks for the latest version available at PyPI."""

    name: str = "cve-ape"
    url: str = f"https://pypi.org/pypi/{name}/json"
    try:
        package_json = requests.get(url).json()
        pypi_version = package_json["info"]["version"]
        if pypi_version != VERSION:
            LOGGER.info(
                f"[bold red]You are running version {VERSION} of {name} but the latest PyPI Version is {pypi_version}.[/]",
                extra={"markup": True},
            )
            if version.parse(VERSION) < version.parse(pypi_version):
                LOGGER.info(
                    "[bold yellow]Alert: We recommend using the latest stable release.[/]",
                    extra={"markup": True},
                )
    except Exception as error:
        LOGGER.warning(
            textwrap.dedent(
                f"""
        -------------------------- Can't check for the latest version ---------------------------
        warning: unable to access 'https://pypi.org/pypi/{name}'
        Exception details: {error}
        Please make sure you have a working internet connection or try again later.
        """
            )
        )
